library (gmp)


iCacheAssociativity <- 16;
iCacheSize <- 4096*1024;
iLocalSize <- 32 * 1024;
iStepSize <- 4096;
iLineSize <- 64;

# The length of the trail followed by the individual pointers.
iLength = iCacheSize / iStepSize;

# The number of sets exercised by the benchmark. This is at
# most the number of sets present in the cache, but can be
# smaller if the length of the trail is smaller.
iSets = min (iCacheSize / iLineSize / iCacheAssociativity, iLength);

iPointersBoth <- 16;
iDistanceMin <- iLocalSize / iLineSize / iPointersBoth;
iDistanceMax <- iLength / iPointersBoth;
stopifnot (iDistanceMin < iDistanceMax);
iDistanceRemote <- round ((iDistanceMax + iDistanceMin) / 2);

# Relative speed of the two cores.
fRate <- 1;


#######################################################
# Helper functions for large precision calculations ...

lBigFactorialCache = NULL;


big.factorial.scalar <- function (iNumber)
{
  xCached <- lBigFactorialCache [[as.character (iNumber)]];

  if (is.null (xCached))
  {
    xResult <- as.bigz (1);
    if (iNumber > 1)
    {
      for (iIndex in 1:iNumber)
      {
        xResult <- xResult * iIndex;
      }
    }

    lNewCacheEntry <- list (xResult);
    attr (lNewCacheEntry, "names") <- as.character (iNumber);
    lBigFactorialCache <<- c (lBigFactorialCache, lNewCacheEntry);

    return (xResult);
  }
  else return (xCached);
}

big.factorial <- function (vNumbers)
{
  # For whatever reason, the list has to be built with one
  # number already present, otherwise it gets converted
  # to incorrect type ...

  if (length (vNumbers))
  {
    vResult <- big.factorial.scalar (vNumbers [1]);
    if (length (vNumbers) > 1)
    {
      for (iNumber in vNumbers [2:length (vNumbers)])
      {
        vResult <- c (vResult, big.factorial.scalar (iNumber));
      }
    }
    return (vResult);
  }
  else return (NULL);
}


big.choose <- function (iN, iK)
{
  return (as.bigq (big.factorial (iN), big.factorial (iN - iK) * big.factorial (iK)));
}


big.sum <- function (vxNumbers)
{
  return (cumsum (vxNumbers) [length (vxNumbers)]);
}


big.power <- function (xNumber, iExponent)
{
  xResult <- as.bigq (numerator (xNumber) ^ iExponent, denominator (xNumber) ^ iExponent);
}


#############################
# Calculations themselves ...

lBigSelectSetsFromGivenProbCache = NULL;

# Calculates probability that a given number of randomly
# picked addresses falls into a given number of sets.
select.sets.from.given.prob.scalar <- function (iAddresses, iUsedSets)
{
  stopifnot (iAddresses >= iUsedSets)
  if (iUsedSets == 0) return (0);

  xCached <- lBigSelectSetsFromGivenProbCache [[paste (iAddresses, iUsedSets)]];

  if (is.null (xCached))
  {
    # The probability that at most the given number of sets is selected
    xOneAddressSelectsSetsProb <- as.bigq (iUsedSets, iSets);
    xAllAddressesSelectSetsProb <- big.power (xOneAddressSelectsSetsProb, iAddresses);

    if (iUsedSets > 1)
    {
      viSubsets <- 1:(iUsedSets-1);

      # The probabilities that a smaller number of sets is selected
      vxAllAddressesSelectSubsetsProb <- select.sets.from.given.prob (iAddresses, viSubsets);

      # The number of subsets that have a smaller number of sets
      vxSubsetCombinations <- big.choose (iUsedSets, viSubsets);

      xSubsetProb <- big.sum (vxAllAddressesSelectSubsetsProb * vxSubsetCombinations);
    }
    else xSubsetProb <- 0;

    xResult <- xAllAddressesSelectSetsProb - xSubsetProb;

    lNewCacheEntry <- list (xResult);
    attr (lNewCacheEntry, "names") <- paste (iAddresses, iUsedSets);
    lBigSelectSetsFromGivenProbCache <<- c (lBigSelectSetsFromGivenProbCache, lNewCacheEntry);

    return (xResult);
  }
  else return (xCached);
}


# A recursive version of the same ...


select.sets.from.given.prob.scalar <- function (iAddresses, iUsedSets)
{
  if (iAddresses == 1 && iUsedSets == 1) return (as.bigq (1, iSets));
  if (iAddresses < iUsedSets) return (0);
  if (iUsedSets == 0) return (0);

  xCached <- lBigSelectSetsFromGivenProbCache [[paste (iAddresses, iUsedSets)]];

  if (is.null (xCached))
  {
    xSameSetsInLessAddressesProb <- select.sets.from.given.prob.scalar (iAddresses - 1, iUsedSets);
    xNewAddressInSameSetsProb <- as.bigq (iUsedSets, iSets);

    xLessSetsInLessAddressesProb <- select.sets.from.given.prob.scalar (iAddresses - 1, iUsedSets - 1);
    xNewAddressInNewSetProb <- as.bigq (1, iSets);
    iNewSetChoices <- iUsedSets;

    xResult = xSameSetsInLessAddressesProb * xNewAddressInSameSetsProb +
              xLessSetsInLessAddressesProb * xNewAddressInNewSetProb * iNewSetChoices;

    lNewCacheEntry <- list (xResult);
    attr (lNewCacheEntry, "names") <- paste (iAddresses, iUsedSets);
    lBigSelectSetsFromGivenProbCache <<- c (lBigSelectSetsFromGivenProbCache, lNewCacheEntry);

    return (xResult);
  }
  else return (xCached);
}


select.sets.from.given.prob <- function (iAddresses, viUsedSetses)
{
  # For whatever reason, the list has to be built with one
  # number already present, otherwise it gets converted
  # to incorrect type ...

  if (length (viUsedSetses))
  {
    vResult <- select.sets.from.given.prob.scalar (iAddresses, viUsedSetses [1]);
    if (length (viUsedSetses) > 1)
    {
      for (iUsedSets in viUsedSetses [2:length (viUsedSetses)])
      {
        vResult <- c (vResult, select.sets.from.given.prob.scalar (iAddresses, iUsedSets));
      }
    }
    return (vResult);
  }
  else return (NULL);
}


# Calculate probability that given number of randomly
# selected addresses falls into a given number of sets.
select.sets.prob <- function (iAddresses, iUsedSets)
{
  # The probability that a given set of the given size is selected.
  xGivenSetsSelectedProb <- select.sets.from.given.prob (iAddresses, iUsedSets);

  # How many combinations of the selection of given sets
  xSetCombinations <- big.choose (iSets, iUsedSets);

  return (xGivenSetsSelectedProb * xSetCombinations);
}


# Calculate probability that given number of accesses in given
# pattern hits a particular set a given number of times.
exactly.hits.prob <- function (iHits, iPointers, iDistance, iAccesses)
{
  stopifnot (iAccesses > 0);

  # Calculate the number of different addresses that get visited.

  # As long as repetition does not happen, every access is new.
  fAddressesBeforeRepetition <- min (iAccesses, iPointers * iDistance);
  # Once repetition happens only the first pointer accesses are new.
  fAddressesDuringRepetition <- min (0, (iAccesses - iPointers * iDistance) / iPointers);
  # The number of different addresses cannot be more than the entire trail.
  fAddresses <- min (iLength, fAddressesBeforeRepetition + fAddressesDuringRepetition);

  # The maximum number of different sets that can be potentially visited.
  fPotentiallyVisitedSets <- min (iSets, fAddresses);

  # We iterate over the number of different sets that are truly visited
  # and calculate the probability that that this number of sets is
  # visited and the probability that the sets contain a set of
  # choice. Then, we use conditional probability to see how
  # likely it is that this set of choice is hit the given
  # number of times.

  viVisitedSets <- 1:fPotentiallyVisitedSets;

  vxSetsVisitedProb <- select.sets.prob (fAddresses, viVisitedSets);
  vfSetsVisitedProb <- as.double (vxSetsVisitedProb);

  vfSetsContainSetOfChoiceProb <- viVisitedSets / iSets;

  if (iHits)
  {
    # We calculate for nonzero hits, we therefore consider the
    # case when the sets do contain the set of choice.

    vfSetHitProb <- (1 / viVisitedSets) ^ iHits;
    vfSetMissProb <- (1 - 1 / viVisitedSets) ^ (iAccesses - iHits);
    iCombinations <- choose (iAccesses, iHits);

    vfExactHitsProb <- vfSetsContainSetOfChoiceProb * vfSetHitProb * vfSetMissProb * iCombinations;
  }
  else
  {
    # We calculate for zero hits, we therefore consider the
    # case when the sets do not contain the set of choice.

    vfExactHitsProb <- 1 - vfSetsContainSetOfChoiceProb;
  }

  return (sum (vfExactHitsProb * vfSetsVisitedProb));
}


at.most.hits.prob <- function (iMaximumHits, iPointers, iDistance, iAccesses)
{
  vfProbability <- NULL;

  for (iExactHits in 0:iMaximumHits)
  {
    vfProbability <- c (vfProbability, exactly.hits.prob (iExactHits, iPointers, iDistance, iAccesses));
  }

  return (sum (vfProbability));
}


at.most.double.hits.prob <- function (iMaximumHits, iPointers, iDistanceOne, iAccessesOne, iDistanceTwo, iAccessesTwo)
{
  # Yeah, quadratic, so what ? :-)

  vfProbability <- NULL;

  for (iExactHits in 0:iMaximumHits)
  {
    for (iHitsOne in 0:iExactHits)
    {
      iHitsTwo <- iExactHits - iHitsOne;
      vfProbability <- c (vfProbability, exactly.hits.prob (iHitsOne, iPointers, iDistanceOne, iAccessesOne) *
                                         exactly.hits.prob (iHitsTwo, iPointers, iDistanceTwo, iAccessesTwo));
    }
  }

  return (sum (vfProbability));
}


access.fault.prob.isolated <- function (iPointers, iDistance)
{
  # The probability of an access fault in the isolated case is
  # the probability that a single set is accessed at least
  # iCacheAssociativity times. Accesses come from one
  # core only.

  return (1 - at.most.hits.prob (iCacheAssociativity - 1, iPointers, iDistance, iPointers * iDistance));
}


access.fault.prob.shared <- function (iPointers, iDistanceOne, iDistanceTwo)
{
  # The probability of an access fault in the shared case is
  # the probability that a single set is accessed at least
  # iCacheAssociativity times. Accesses come from two
  # cores.

  return (1 - at.most.double.hits.prob (iCacheAssociativity - 1, iPointers, iDistanceOne, iPointers * iDistanceOne,
                                                                            iDistanceTwo, iPointers * iDistanceOne * fRate));
}


vectorizing.access.fault.prob.isolated <- function (vDistance)
{
  return (sapply (vDistance, function (x) access.fault.prob.isolated (iPointersBoth, x)));
}

vectorizing.access.fault.prob.shared <- function (vDistance)
{
  return (sapply (vDistance, function (x) access.fault.prob.shared (iPointersBoth, x, iDistanceRemote)));
}


curve (vectorizing.access.fault.prob.shared, iDistanceMin, iDistanceMax, min (16, iDistanceMax - iDistanceMin), type="p", add=FALSE);
curve (vectorizing.access.fault.prob.isolated, iDistanceMin, iDistanceMax, min (16, iDistanceMax - iDistanceMin), type="l", add=TRUE);

